"""
Execution on various engines.
"""
from dslibrary.run_models import ModelRunner


class ModelEngine(object):
    def get_metadata(self, uri: str):
        """
        Obtain metadata for the given model or pipeline step.  This metadata includes input and output schemas, which
        will help us know how we can connect the steps of a pipeline, as well as parameter names.
        """

    # TODO enumerate models?
    # TODO wrapper for container build process?
    # TODO compare to SageMaker, AzureML, GoogleAI

    def run(self, model_run_spec: ModelRunner):
        """
        Execute the given model.
        """


class ModelEngineLocal(ModelEngine):
    """
    Execute models locally.
    """
    def __init__(self, default_folder: str=None):
        """
        :param default_folder:  A folder containing model folders.  A model is referenced by supplying the name
                                of a subfolder as its URI.
        """
        self.default_folder = default_folder

    def get_metadata(self, model_uri: str):
        pass

    def run(self, model_run_spec: ModelRunner):
        # TODO with a path we can invoke as a subprocess, but with a method name we can invoke more quickly
        #  - how about we add a method name to the entry point metadata?  (package.module.method)
        #    - if it is a module with a __main__, we invoke that, if it is a method, we call that
        pass




'''
def get_metadata(self):

ModelEngineMLFlow() - the URI is used to invoke the model
ModelEngineKubernetesJob() - etc.
ModelEngineSageMaker() - etc.

engine = ModelEngineLocal(default_model_path="/my_models")
run = ModelRunner("model1")
run.set_input(...)
engine.run(run)
'''




import unittest
from dslibrary.utils import dbconn


class TestDBConn(unittest.TestCase):

    def test__is_read_operation(self):
        assert dbconn._is_read_operation("\nselect\n  something") is True
        assert dbconn._is_read_operation("insert whatever") is False
        assert dbconn._is_read_operation("Select x") is True
        assert dbconn._is_read_operation("show databases") is True

    def test__is_table_name(self):
        for v in ["", "select something", "a b", "`table` x", "4"]:
            assert dbconn._is_table_name(v) is None
        for v, expect in {"x": "x", "_x": "_x", "`x`": "x", "x4": "x4", '"x"': "x", "[x]": "x", "_": "_", "` `": " "}.items():
            assert dbconn._is_table_name(v) == expect

    def test_read(self):
        log = []
        cols = ("c1", "c2")
        rows = [(1, 2), (10, 20), (100, 200), (1000, 2000)]
        def read(operation, parameters):
            log.append((operation, parameters))
            return tuple([col] + [None]*5 for col in cols), rows, None
        cursor = dbconn.Connection(read).cursor()
        cursor.execute("select %s, %s", ("a", "b"))
        assert log[0] == ("select %s, %s", ["a", "b"])
        descr = cursor.description
        assert descr[0][0] == cols[0]
        assert descr[1][0] == cols[1]
        assert cursor.rowcount == 4
        assert cursor.fetchone() == rows[0]
        assert cursor.fetchmany(2) == [rows[1], rows[2]]
        assert cursor.fetchall() == [rows[3]]
        assert cursor.fetchone() is None
        assert cursor.fetchmany(1) == []
        assert cursor.fetchall() == []
        # table only
        log.clear()
        cursor.execute("table")
        assert log == [('select * from "table"', [])]
        descr = cursor.description
        assert descr[0][0] == cols[0]
        assert descr[1][0] == cols[1]
        assert cursor.fetchall() == rows

    def test_read_mult(self):
        log = []
        cols = ("c1", "c2")
        rows = [(1, 2), (10, 20)]
        def read(operation, parameters):
            log.append((operation, parameters))
            return tuple([col] + [None]*5 for col in cols), rows, None
        cursor = dbconn.Connection(read).cursor()
        cursor.execute("select %s; select %s", ("a", "b"))
        assert len(log) == 1
        assert log[0] == ("select %s", ["a"])
        assert cursor.fetchall() == rows
        cursor.nextset()
        assert len(log) == 2
        assert log[1] == ("select %s", ["b"])
        assert cursor.fetchall() == rows
        self.assertRaises(Exception, lambda: cursor.nextset())

    def test_write(self):
        log = []
        def read(operation, parameters):
            raise Exception()
        def write(operation, parameters):
            log.append((operation, parameters))
        cursor = dbconn.Connection(read, write).cursor()
        cursor.execute("update 1")
        assert log[0] == ('update 1', None)
        log.clear()
        cursor.executemany("update %s", [(1,), (2,)])
        assert log[0] == ('update %s', (1,))
        assert log[1] == ('update %s', (2,))

    def test_read_quotes(self):
        log = []
        cols = ("c1", "c2")
        rows = [(1, 2), (10, 20), (100, 200), (1000, 2000)]
        def read(operation, parameters):
            log.append((operation, parameters))
            return tuple([col] + [None]*5 for col in cols), rows, None
        cursor = dbconn.Connection(read).cursor()
        cursor.execute('select * from "quoted"')
        assert log[0] == ('select * from "quoted"', [])

    def test_read_more(self):
        cols = ("c1",)
        for mode in ["fetchone", "fetchall", "fetchmany"]:
            rows = [(_,) for _ in range(100)]
            def chunk():
                out = []
                for _ in range(30):
                    if not rows:
                        break
                    out.append(rows.pop(0))
                return out

            def read(operation, parameters):
                out = chunk()
                return cols, out, 1
            def read_more(more):
                out = chunk()
                return out, 1 if rows else None
            cursor = dbconn.Connection(read, read_more=read_more).cursor()
            cursor.execute('select ...')
            if mode == "fetchone":
                for _ in range(100):
                    assert cursor.fetchone() == (_,)
                assert cursor.fetchone() is None
            elif mode == "fetchmany":
                batch = cursor.fetchmany(25)
                assert batch == [(_,) for _ in range(25)]
                batch = cursor.fetchmany(100)
                assert batch == [(_,) for _ in range(25, 100)]
                assert cursor.fetchmany(100) == []
            elif mode == "fetchall":
                batch = cursor.fetchall()
                assert batch == [(_,) for _ in range(100)]
                assert cursor.fetchall() == []

    # TODO it doesn't many any sense to emulate sqlite - remove all of this code
    '''
    def test_to_sql(self):
        log = []
        df = pandas.DataFrame({"x": [1, 2]})
        def read(operation, parameters):
            if "sqlite_master" in operation:
                return ["name"], [], None
            raise Exception()
        def write(operation, parameters):
            log.append((operation, parameters))
        conn = dbconn.Connection(read, write)
        df.to_sql("TABLE", conn)
        assert "CREATE TABLE" in log[0][0]
        assert "CREATE INDEX" in log[1][0]
        assert "INSERT INTO" in log[2][0]
        assert "INSERT INTO" in log[3][0]
    '''

    def test_misc(self):
        cols = ("c1", "c2")
        rows = [(1, 2), (10, 20), (100, 200), (1000, 2000)]
        def read(operation, *parameters):
            return tuple([col] + [None]*5 for col in cols), rows, None
        cursor = dbconn.Connection(read).cursor()
        rows = list(cursor.execute("select * from X"))
        cols = [v[0] for v in cursor.description]
        assert rows == [(1, 2), (10, 20), (100, 200), (1000, 2000)]
        assert cols == ["c1", "c2"]

    def test_limited_cache_size(self):
        """
        Rows would pile up forever if we didn't throw them out.
        """
        cols = ("x", None, None, None, None, None)
        def read(operation, *parameters):
            return cols, [[_] for _ in range(2000)], None
        cursor = dbconn.Connection(read).cursor()
        cursor.execute("whatever")
        rows = cursor.fetchmany(1500)
        assert rows == [[_] for _ in range(1500)]
        rows = cursor.fetchmany(100)
        assert rows == [[_] for _ in range(1500, 1600)]
        assert len(cursor._rows) == 500, "here we see the cache has been truncated"
        rows = cursor.fetchall()
        assert rows == [[_] for _ in range(1600, 2000)]




import unittest
import mock
import io
import os
import numpy
import pandas
import shutil
import tempfile
import yaml

from dslibrary import DSLibrary
from dslibrary.front import EVALUATION_RESULT_ALIAS, METRICS_ALIAS, DSLibraryException
from dslibrary.metadata import Metadata
from dslibrary.transport.to_local import DSLibraryLocal


class TestFront(unittest.TestCase):

    def test_metadata(self):
        """
        The base class doesn't know how to get metadata so it returns a null instance.
        """
        dsl = DSLibrary()
        m = dsl.get_metadata()
        assert isinstance(m, Metadata)
        assert m.uri == ""
        assert m.entry_points == {}

    def test_open_resource__input_mapping(self):
        """
        Options for input come from mapping values, which override supplied values.
        """
        log = []
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                log.append((path, mode, kwargs))
                return "H"
        MyInst(spec={"inputs": {"a": {"uri": "aaa", "option1": 1, "option3": 3}}}).open_resource("a", option1=2, option2=22)
        assert log[0] == ('aaa', 'rb', {'option1': 1, 'option2': 22, 'option3': 3}), log[0]

    def test_open_resource__output_mapping(self):
        """
        Options for output come from mapping values, which override supplied values.
        """
        log = []
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                log.append((path, mode, kwargs))
                return "H"
        MyInst(spec={"outputs": {"a": {"uri": "aaa", "option1": 1, "option3": 3}}}).open_resource("a", 'w', option1=2, option2=22)
        self.assertEqual(log[0], ('aaa', 'w', {'option1': 1, 'option2': 22, 'option3': 3}))

    def test_open_resource__bypass_mapping(self):
        """
        Most inputs and outputs are mapped, but you can bypass mapping and specify a URI to open or a local file.
        """
        log = []
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                log.append((path, mode, kwargs))
                return "H"
        MyInst().open_resource("s3://bucket/path", open_option_1=1)
        assert log[0] == ('s3://bucket/path', 'rb', {'open_option_1': 1})
        MyInst().open_resource("./local/file")
        assert log[1] == ('./local/file', 'rb', {})

    def test_open_model_binary(self):
        """
        The default implementation just assumes certain filenames for the 'model-binary' data.
        """
        log = []
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                log.append((path, mode, kwargs))
                return "H"
        MyInst().open_model_binary()
        assert log[0] == ('model-binary', 'rb', {})
        MyInst().open_model_binary("part1")
        assert log[1] == ('model-binary/part1', 'rb', {})

    def test_set_evaluation_result(self):
        """
        Model evaluation code can use this hook to report whether the model passed or failed.
        """
        log = []
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                buf = io.StringIO()
                buf.close = lambda: None
                log.append((path, buf))
                return buf
        MyInst().set_evaluation_result(True)
        MyInst().set_evaluation_result(False, reason="because")
        assert log[0][0] == log[1][0] == EVALUATION_RESULT_ALIAS
        r = log[0][1].getvalue()
        assert r == '{"uri": "", "success": true}\n', r
        r = log[1][1].getvalue()
        assert r == '{"uri": "", "success": false, "reason": "because"}\n', r

    def test_get_sql_connection__mapping(self):
        """
        The normal use case is have the caller supply all the connection information.
        """
        dsl = DSLibrary(spec={"inputs": {"db": {"uri": "mysql://host/db", "username": "u"}}})
        def connect(user, host, database, **kwargs):
            assert user == "u"
            assert host == "host"
            assert database == "db"
            return "CONN"
        with mock.patch("pymysql.connect", connect):
            r = dsl.get_sql_connection("db")
            assert r == "CONN"

    def test_write_resources__dataframes_and_series(self):
        """
        Several types are supported for columnar data.
        """
        log = []
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                buf = io.StringIO()
                buf.close = lambda: None
                log.append((path, buf))
                return buf
        dsl = MyInst()
        # series
        dsl.write_resource("x", pandas.Series([1, 2, 3]))
        assert log[0][0] == "x"
        self.assertEqual(log[0][1].getvalue(), 'x\n1\n2\n3\n')
        log.clear()
        # numpy array
        dsl.write_resource("x", numpy.array([1, 2, 3]))
        assert log[0][0] == "x"
        self.assertEqual(log[0][1].getvalue(), 'x\n1\n2\n3\n')
        log.clear()
        # dataframe
        dsl.write_resource("x", pandas.DataFrame({"y": [1, 2, 3]}))
        assert log[0][0] == "x"
        self.assertEqual(log[0][1].getvalue(), 'y\n1\n2\n3\n')
        log.clear()

    def test_default_metrics_output(self):
        project = tempfile.mkdtemp()
        dsl = DSLibraryLocal(project)
        dsl.log_metric("x", 1)
        r = dsl.get_last_metric("x")
        assert r.value == 1
        # verify it was written as JSON
        fn = os.path.join(project, METRICS_ALIAS)
        assert os.path.exists(fn)
        with open(fn, 'r') as f_r:
            assert f_r.read().startswith("{")
        shutil.rmtree(project)

    def test_alt_metrics_output1(self):
        project = tempfile.mkdtemp()
        dsl = DSLibraryLocal(project, spec={"outputs": {METRICS_ALIAS: {"format": "csv"}}})
        dsl.log_metric("x", 1)
        # verify it was written as CSV
        fn = os.path.join(project, METRICS_ALIAS)
        assert os.path.exists(fn)
        with open(fn, 'r') as f_r:
            assert f_r.read().startswith("uri,")
        shutil.rmtree(project)

    def test_alt_metrics_output2(self):
        project = tempfile.mkdtemp()
        dsl = DSLibraryLocal(project, spec={"outputs": {METRICS_ALIAS: {"uri": "metrics.csv"}}})
        dsl.log_metric("x", 1)
        # verify it was written as CSV
        fn = os.path.join(project, "metrics.csv")
        assert os.path.exists(fn)
        with open(fn, 'r') as f_r:
            assert f_r.read().startswith("uri,")
        shutil.rmtree(project)

    def test_model_pickling(self):
        project = tempfile.mkdtemp()
        dsl = DSLibraryLocal(project, spec={"outputs": {METRICS_ALIAS: {"uri": "metrics.csv"}}})
        # save and restore
        my_model = {"x": 1}
        dsl.save_pickled_model(my_model)
        restored = dsl.load_pickled_model()
        assert restored == my_model
        # verify local storage
        assert os.path.exists(project + "/model-binary")
        shutil.rmtree(project)

    def test_open_resource__default_uri(self):
        """
        Specify a name and a default URI, for cleaner overriding.
        """
        log = []
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                log.append((path, kwargs))
                return "H"
        MyInst().open_resource("x", uri="default")
        MyInst(spec={"inputs": {"x": {"uri": "override", "option1": 1}}}).open_resource("x", uri="default")
        self.assertEqual(log, [('default', {}), ('override', {'option1': 1})])

    def test_load_dataframe__default_uri(self):
        log = []
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                log.append(path)
                return io.BytesIO(b"x\n1\n2")
        df = MyInst().load_dataframe("x", uri="default.csv")
        df = MyInst(spec={"inputs": {"x": {"uri": "override.csv"}}}).load_dataframe("x", uri="default.csv")
        self.assertEqual(log[0], 'default.csv')
        self.assertEqual(log[-1], 'override.csv')

    def test_load_dataframe__from_sql(self):
        log = []
        class MyInst(DSLibrary):
            def get_sql_connection(self, resource_name: str, for_write: bool=False, **kwargs):
                class DbCursor(object):
                    description = [("x", None)]
                    def execute(self, sql):
                        log.append(sql)
                    def __iter__(self):
                        return iter([(1,), (2,)])
                class DbConn(object):
                    def cursor(self):
                        return DbCursor()
                    def close(self):
                        log.append("close")
                return DbConn()
        df = MyInst().load_dataframe("x", sql_table="table1")
        self.assertEqual(list(df.x), [1, 2])
        self.assertEqual(log, ['SELECT * from table1', 'close'])

    def test_load_dataframe__from_nosql(self):
        class MyInst(DSLibrary):
            def get_nosql_connection(self, resource_name: str, for_write: bool=False, **kwargs):
                class MyNoSql(object):
                    def query(self, collection, **kwargs):
                        assert collection == "table1"
                        return [{"x": 1}, {"x": 2}]
                return MyNoSql()
        df = MyInst().load_dataframe("x", nosql_collection="table1")
        self.assertEqual(list(df.x), [1, 2])

    def test_load_dataframe__sql__custom_open_args(self):
        """
        Custom arguments can be passed through to get_sql_connection().
        """
        class MyInst(DSLibrary):
            def get_sql_connection(self, resource_name: str, database: str=None, for_write: bool=False, **kwargs):
                assert resource_name == "engine"
                assert database == "db"
                assert kwargs["custom1"] == 123
                class DbCursor(object):
                    description = [("x", None)]
                    def execute(self, sql):
                        pass
                    def __iter__(self):
                        return iter([])
                class DbConn(object):
                    def cursor(self):
                        return DbCursor()
                    def close(self):
                        pass
                return DbConn()
        MyInst().load_dataframe("engine", sql_table="table1", database="db", custom1=123)

    def test_open_resource__named_filesystem(self):
        """
        A resource can be opened through a specified filesystem engine/provider.
        """
        class MyInst(DSLibrary):
            def get_filesystem_connection(self, resource_name: str, for_write: bool=False, **kwargs):
                assert resource_name == "fs"
                assert kwargs.get("custom1") == 111
                class FS(object):
                    def open(self, path, mode):
                        assert path == "path"
                        assert mode == "r"
                        return "FH"
                return FS()
        r = MyInst().open_resource("path", mode="r", filesystem="fs", custom1=111)
        assert r == "FH"

    def test_load_dataframe__named_filesystem(self):
        """
        A dataframe can be opened through a specified filesystem engine/provider.
        """
        class MyInst(DSLibrary):
            def get_filesystem_connection(self, resource_name: str, for_write: bool=False, **kwargs):
                assert resource_name == "fs"
                class FS(object):
                    def open(self, path, mode):
                        assert path == "path"
                        assert mode == "rb"
                        return io.BytesIO(b"x\n1\n2\n3")
                return FS()
        df = MyInst().load_dataframe("path", filesystem="fs", format="csv")
        assert list(df.x) == [1, 2, 3]

    def test_rw_run_data(self):
        writes = {}
        class MyInst(DSLibrary):
            def open_run_data(self, filename: str, mode: str='rb'):
                if mode == 'rb':
                    v = writes[filename].getvalue()
                    return io.BytesIO(v)
                wr = io.BytesIO()
                writes[filename] = wr
                wr.close = lambda: None
                return wr
        dsl = MyInst()
        df = pandas.DataFrame({"x": [1, 2]})
        dsl.write_run_data("x.csv", df)
        self.assertEqual(writes["x.csv"].getvalue(), b'x\n1\n2\n')
        df2 = dsl.load_dataframe("x.csv", run_data=True)
        self.assertEqual(list(df2.x), list(df.x))
        self.assertEqual(dsl.read_run_data("x.csv"), b'x\n1\n2\n')

    def test_open_args_passed_through_to_opener(self):
        log = []
        inst = self
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                log.append(path)
                inst.assertEqual(kwargs, {'arg1': 123})
                return io.BytesIO(b"abc")
        # just a filename
        fh = MyInst().open_resource("x", arg1=123, format="xyz")
        assert fh.read() == b"abc"
        # uri overrides name
        fh = MyInst().open_resource("x", uri="default.csv", arg1=123, format="xyz")
        assert fh.read() == b"abc"
        # input mapping maps 'x' to something else
        fh = MyInst(spec={"inputs": {"x": {"uri": "override.csv"}}}).open_resource("x", uri="default.csv", arg1=123, format="xyz")
        assert fh.read() == b"abc"
        # verify filenames
        self.assertEqual(log, ['x', 'default.csv', 'override.csv'])

    def test_setup_code_paths(self):
        """
        You can cause 'sys.path' to be extended to include particular folders.
        This is normally done using an environment variable.  Here we are just testing the method that applies the
        changes.
        """
        inst = DSLibrary()
        inst._spec = {"code_paths": ["a", "b"]}
        mock_paths = []
        with mock.patch("sys.path", mock_paths):
            inst._setup_code_paths()
        self.assertEqual(mock_paths, ["a", "b"])

    def test_abstract_base_methods(self):
        """
        The methods that have to be filled in raise exceptions when called.
        """
        inst = DSLibrary()
        self.assertRaises(DSLibraryException, lambda: inst._opener("path", mode='r'))
        self.assertRaises(DSLibraryException, lambda: inst.open_run_data("fn"))

    def test_get_filesystem_connection(self):
        """
        The default filesystem accessor.
        """
        inst = DSLibrary()
        log = []
        with mock.patch("dslibrary.front.connect_to_filesystem", lambda **k: log.append(k) or "FS"):
            fs = inst.get_filesystem_connection("resource", for_write=True, arg1=123)
            self.assertEqual(fs, "FS")
        self.assertEqual(log, [{'uri': 'resource', 'for_write': True, 'arg1': 123}])

    def test_get_sql_connection(self):
        """
        The default SQL accessor.
        """
        inst = DSLibrary()
        log = []
        with mock.patch("dslibrary.front.connect_to_database", lambda **k: log.append(k) or "NoSqlConn"):
            fs = inst.get_sql_connection("resource", for_write=True, arg1=123)
            self.assertEqual(fs, "NoSqlConn")
        self.assertEqual(log, [{'uri': 'resource', 'library': None, 'for_write': True, 'arg1': 123}])

    def test_get_nosql_connection(self):
        """
        The default NoSQL accessor.
        """
        inst = DSLibrary()
        log = []
        with mock.patch("dslibrary.front.connect_to_nosql", lambda **k: log.append(k) or "SqlConn"):
            fs = inst.get_nosql_connection("resource", for_write=True, arg1=123)
            self.assertEqual(fs, "SqlConn")
        self.assertEqual(log, [{'uri': 'resource', 'library': None, 'for_write': True, 'arg1': 123}])

    def test_log_mlflow(self):
        inst = DSLibrary()
        inst._mlflow_all = True
        log = []
        with mock.patch("mlflow.log_param", lambda k, v: log.append((k, v))):
            inst.log_param("x", 1)
        self.assertEqual(log, [("x", 1)])
        log.clear()
        with mock.patch("mlflow.log_metric", lambda *a: log.append(a)):
            inst.log_metric("q", 222)
            inst.log_metrics({"z": 9}, step=2)
        self.assertEqual(log, [('q', 222, 0), ('z', 9, 2)])
        log.clear()
        with mock.patch("mlflow.log_dict", lambda *a: log.append(a)):
            inst.log_dict({"x": 1}, "f.json")
        self.assertEqual(log, [({'x': 1}, 'f.json')])
        log.clear()
        with mock.patch("mlflow.log_artifact", lambda *a: log.append(a)):
            inst.log_artifact("f_local", "f_store")
        self.assertEqual(log, [('f_local', 'f_store')])
        log.clear()
        # not a log function but close enough
        self.assertRaises(DSLibraryException, lambda: inst.get_metrics("x"))

    def test_mlflow_start_end(self):
        inst = DSLibrary()
        inst._mlflow_all = True
        log = []
        with mock.patch("mlflow.mlflow.start_run", lambda: log.append("start")):
            with mock.patch("mlflow.mlflow.end_run", lambda: log.append("end")):
                with mock.patch("mlflow.active_run", lambda: log.append("active_run")):
                    # TODO mock this: mlflow.active_run().info.run_id
                    with inst.start_run():
                        pass
        self.assertEqual(log, ['start', 'active_run', 'end'])

    def test_yaml(self):
        data = [
            {"age": 41, "name": {"first": "John", "last": "Smith"}}
        ]
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                return io.StringIO(yaml.dump(data))
        df = MyInst().load_dataframe("data.yaml")
        assert list(df.age) == [41]
        assert list(df["name.first"]) == ["John"]

    def test_xlsx(self):
        fn = os.path.join(os.path.dirname(__file__), "test_data/two_cols.xlsx")
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                return open(path, mode)
        df = MyInst().load_dataframe(fn)
        self.assertEqual(list(df.a), [2, 7])
        self.assertEqual(list(df.b), [5, 9])

    def test_xlsx__unseekable_stream(self):
        """
        The xlsx parser requires a seekable stream, so when we supply an unseekable stream
        it does some special processing.
        """
        fn = os.path.join(os.path.dirname(__file__), "test_data/two_cols.xlsx")
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                with open(path, mode) as f_r:
                    data = f_r.read()
                class Stream(io.RawIOBase):
                    def __init__(self, data):
                        self.data = data
                    def read(self, n=None):
                        if n is None:
                            n = len(self.data)
                        out = self.data[:n]
                        self.data = self.data[n:]
                        return out
                    def readable(self):
                        return True
                    def seekable(self):
                        return False
                    def __enter__(self):
                        return self
                    def __exit__(self, exc_type, exc_val, exc_tb):
                        pass
                return Stream(data)
        df = MyInst().load_dataframe(fn)
        self.assertEqual(list(df.a), [2, 7])
        self.assertEqual(list(df.b), [5, 9])



"""
Format detection.
"""
import typing
import os
import re
import csv
from collections import namedtuple, defaultdict

FileFormatDescription = namedtuple("FileFormatDescription", "format read_args size avg_line_size")

# values that are unlikely to be header names, i.e. blank or numeric
PTN_NOT_HEADER = re.compile(r'^([-+]?(\d+(.\d*)?|.\d+)([eE][-+]?\d+)?)?$')


def detect_format(
        filename: str, sampler: typing.Union[typing.Callable[[], typing.Union[bytes, bytearray]], bytes, str, bytearray, None]=None,
        sizer: typing.Union[typing.Callable[[], int], typing.Union[int, None]]=None, supplied_options: dict=None,
        fallback_to_text: bool=False
) -> FileFormatDescription:
    """
    General purpose format detector / sniffer.

    :param filename:        Name or URL for file, used specifically and only to guess format from extension.
    :param sampler:         A method that returns a sample, a sample, or None.
    :param sizer:           A method that returns the file's size, the size, or None.
    :param supplied_options:    Caller-supplied formatting options which will override the detection process.
    :param fallback_to_text:  True to assume any unrecognized format that looks like a text file should be read line by line.

    :returns:  A FileFormatDescription, with fields:
                    format              File format: csv, json, ...
                    read_args           Arguments compatible with pandas.read_***() methods.
                    size                File size or None.
                    avg_line_size       Estimated mean size of lines, which can be useful for choosing a processing
                                        technique.
    """
    # determine format of file
    ext = find_url_extension(filename)
    format = ext
    if supplied_options and "format" in supplied_options:
        format = supplied_options["format"]
    read_args = {}
    # normalize the format and infer things
    FORMAT_ALIASES_ETC = {
        "csv": ("csv", {}),
        "json": ("json", {}),
        "yml": ("yaml", {}),
        "tab": ("csv", {"delimiter": "\t"}),
        "tsv": ("csv", {"delimiter": "\t"}),
    }
    if format:
        format, implied_defaults = FORMAT_ALIASES_ETC.get(format.lower(), (format, {}))    # determine format based on a sample
        read_args.update(implied_defaults)
    size = sizer() if sizer and hasattr(sizer, "__call__") else sizer
    avg_line_size = None
    # use sample of data if available
    # - some formats are not sniffed so we don't sample them
    if sampler and format not in ("yaml", "xlsx", "xls", "parquet"):
        sample = sampler() if hasattr(sampler, "__call__") else sampler
        if not format and looks_like_json(sample, size):
            format = "json"
        elif not format and looks_like_csv(sample):
            format = "csv"
        if format == "csv":
            apply_csv_read_args(sample, read_args)
        elif format == "json":
            apply_json_read_args(sample, read_args)
        elif format in ("yaml", "xlsx", "xls", "parquet"):
            # no 'fallback-to-text' for any recognized format
            pass
        elif fallback_to_text and looks_like_text_file(sample):
            # a single column called 'line'
            format = "csv"
            read_args.update({"delimiter": '\n', "names": ["line"]})
        # calculate average line size
        avg_line_size = average_line_size(sample)
    # override determined options with supplied values
    if supplied_options:
        overrides = dict(supplied_options)
        overrides.pop("format", None)
        read_args.update(overrides)
    # return final synopsis
    return FileFormatDescription(format, read_args, size, avg_line_size)


def find_url_extension(url: str) -> str:
    """
    Find the extension of a filename or URL.
    """
    url = url or ""
    if "#" in url:
        url = url.split("#")[0]
    if "?" in url:
        url = url.split("?")[0]
    return os.path.splitext(url)[1].lower().strip(".")


def average_line_size(sample):
    """
    Calculate mean size of whole lines in sample.
    """
    if not sample:
        return
    line_sep = "\n" if isinstance(sample, str) else b"\n"
    lines = sample.split(line_sep)
    if len(lines) == 1:
        # no linefeeds at all >> guess that it is somewhat more than our sample size
        return len(sample) * 3 / 2 + 1
    n_lines = len(lines) - 1
    total_size = len(line_sep.join(lines[:-1]))
    return int(total_size + n_lines//2) // n_lines


def looks_like_csv(sample):
    """
    Detect fairly obvious CSV format.
    :returns:  Separator or None.
    """
    if not sample:
        return False
    if not isinstance(sample, str):
        try:
            sample = sample.decode("utf-8")
        except UnicodeDecodeError:
            # non-unicode >> assume not CSV
            return False
    lines = sample.split("\n")[:10]
    for splitter, sep, threshold in ((re.compile(r"\t"), "\t", 2), (re.compile(r","), ",", 2), (re.compile(r"\s+"), "\\s+", 3)):
        n_cols = []
        for line in lines:
            if line:
                n_cols.append(len(splitter.split(line)))
        if min(n_cols) != max(n_cols) or n_cols[0] < threshold:
            continue
        return sep


def looks_like_json(sample, size: int=None):
    """
    Detect fairly obvious JSON format.
    """
    if not sample:
        return False
    if not isinstance(sample, str):
        try:
            sample = sample.decode("utf-8")
        except UnicodeDecodeError:
            # non-unicode >> assume not JSON
            return False
    sample = sample.strip()
    # JSON would always start with one of these
    if not sample.startswith("{") and not sample.startswith("["):
        return False
    # JSON cannot contain non-ascii characters
    if re.search(r'[^\x20-\x7F]', sample):
        return False
    # if we know the size and our sample goes to the end we can check that as well
    if size and len(sample) == size:
        if not sample.endswith("}") and not sample.endswith("]"):
            return False
    # looks like JSON, although it could still be a CSV with bracketed columns like "[col1], [col2]" for which
    #   we have an incomplete sample
    return True


def looks_like_text_file(sample):
    """
    Detect a text file, i.e. any kind of file that is not binary.  Such files can be considered, as a fallback, to
    have a single column and no header.
    """
    if not isinstance(sample, str):
        try:
            sample = sample.decode("utf-8")
        except UnicodeDecodeError:
            # non-unicode >> assume not a text file
            return False
    if len(sample) > 10000 and "\n" not in sample:
        return False
    return True


def apply_csv_read_args(sample, read_args):
    """
    Analyze format of a CSV file.
    """
    if not sample:
        return
    if isinstance(sample, (bytes, bytearray)):
        sample = sample.decode("utf-8", errors="ignore")
    sniffer = csv.Sniffer()
    try:
        dialect = sniffer.sniff(sample)
    except csv.Error:
        if not alt_delimiter_detection(sample, read_args):
            read_args["delimiter"] = "\n"
            detect_headerless_csv(sample, read_args)
        return
    # leaving out: "lineterminator"
    for name, dflt in {"delimiter": None, "doublequote": None, "escapechar": None, "quotechar": '"', "quoting": 0,
              "skipinitialspace": False}.items():
        v = getattr(dialect, name)
        if v is not dflt:
            read_args[name] = v
    if read_args.get("delimiter") == " ":
        del read_args["delimiter"]
        read_args["delim_whitespace"] = True
    # special case: no delimiters at all
    detect_headerless_csv(sample, read_args)


def apply_json_read_args(sample, read_args):
    """
    Analyze format of JSON data.
    """
    if not sample:
        return
    if isinstance(sample, (bytes, bytearray)):
        sample = sample.decode("utf-8", errors="ignore")
    partial_line = not sample.endswith("\n")
    lines = [line.strip() for line in sample.strip().split("\n")]
    if partial_line and len(lines) > 2:
        lines = lines[:-1]
    if len(lines) < 2:
        return
    # single JSON object or array per line
    if all(line.startswith("{") and line.endswith("}") for line in lines):
        read_args["lines"] = True
    if all(line.startswith("[") and line.endswith("[") for line in lines):
        read_args["lines"] = True


def alt_delimiter_detection(sample, read_args):
    """
    Sniffer doesn't always detect delimiters very well.
    """
    per_delim = defaultdict(list)
    delims = [(ptn, re.compile(ptn)) for ptn in (",", ";", "\t", r"\s+", "\\|")]
    lines = sample.split("\n")
    for line in lines:
        for delim, delim_ptn in delims:
            n_cols = len(delim_ptn.split(line))
            if n_cols > 1:
                per_delim[delim].append(n_cols)
    rankings = []
    import numpy
    for delim, counts in per_delim.items():
        if len(counts) < len(lines)/2:
            continue
        rankings.append((
            -numpy.std(counts),
            -len(counts),
            delim
        ))
    rankings.sort()
    if rankings:
        read_args["delimiter"] = rankings[0][2]
        return True
    return False


def detect_headerless_csv(sample, read_args):
    """
    Detect a CSV file with no header.
    """
    first_line = sample.split("\n")[0]
    delim = read_args.get("delimiter", ' ')
    if delim == ' ':
        cols = re.split(r'\s+', first_line)
    else:
        cols = first_line.split(delim)
    if all(map(lambda col: PTN_NOT_HEADER.match(col) is not None, cols)):
        read_args["header"] = None
        read_args["names"] = ["col_%d" % (n+1) for n in range(len(cols))]



import unittest
import os
import json
import mock
import tempfile
import pandas
import shutil

import dslibrary
from dslibrary.front import PARAMS_ALIAS, METRICS_ALIAS
from dslibrary.run_models import ModelRunner
from tests.t_utils import reset_env


class TestRunMethod(unittest.TestCase):
    """
    The ModelRunner.run_method() method lets you call your model's main method directly, for development, testing or
    hyper-parameter tuning.
    """

    def test_run_method__intended_approach(self):
        """
        A dslibrary instance is passed to the code you want to run.
        """
        def my_model(dsl):
            assert dsl.get_parameter("x") == 123
        runner = ModelRunner()
        runner.set_parameter("x", 123)
        runner.run_method(my_model)

    def test_run_method__alternate_approach(self):
        """
        The model can also call dslibrary.instance()
        """
        def my_model():
            dsl = dslibrary.instance()
            assert dsl.get_parameter("x") == 123
        runner = ModelRunner()
        runner.set_parameter("x", 123)
        runner.run_method(my_model)

    def test_run_method__params_to_method(self):
        """
        Or the parameters can just be sent to the method.
        """
        def my_model(dsl, x):
            assert x == 123
        runner = ModelRunner()
        runner.set_parameter("x", 123)
        runner.run_method(my_model)

    def test_params_to_method__coerce_type(self):
        """
        Type annotation is applied.
        """
        def my_model(dsl, x: int):
            assert x == 12
        runner = ModelRunner()
        runner.set_parameter("x", "12")
        runner.run_method(my_model)

    def test_kwargs_to_method(self):
        """
        Method can request that all unspecified parameters be sent in kwargs.
        """
        def my_model(dsl, x, **kwargs):
            assert x == 123
            assert kwargs["y"] == 222
        runner = ModelRunner()
        runner.set_parameter("x", 123)
        runner.set_parameter("y", 222)
        runner.run_method(my_model)

    def test_read_dataframe_from_specified_file_with_format(self):
        """
        The model only knows there is an input called 'input1'.  It requests that it be loaded into a dataframe.
        """
        runner = ModelRunner()
        runner.set_input("input1", os.path.dirname(__file__) + "/test_data/test1.csv", format_options={"delim_whitespace": True})
        def my_model(dsl):
            df = dsl.load_dataframe("input1")
            assert list(df.columns) == ["a", "b"], list(df.columns)
            assert list(df.a) == [1, 2]
            assert list(df.b) == [2, 4]
        runner.run_method(my_model)

    def test_write_dataframe_to_specified_file_with_format(self):
        """
        The model has an output called 'output1', and doesn't want to care where that output data should go or what
        format to use.
        """
        with tempfile.NamedTemporaryFile(suffix=".csv") as f_tmp:
            runner = ModelRunner(uri="uri", entry_point="main")
            runner.set_output("output1", f_tmp.name, format_options={"sep": "\t"})
            def my_model(dsl):
                df = pandas.DataFrame({"x": [1, 2], "y": [3, 4]})
                dsl.write_resource("output1", df)
            runner.run_method(my_model)
            df = pandas.read_csv(f_tmp.name, sep='\t')
            assert list(df.columns) == ["x", "y"]
            assert list(df.x) == [1, 2]
            assert list(df.y) == [3, 4]

    def test_model_parameter_misc(self):
        """
        The model declares the type of its parameters, and such things as default values, validation and coercion
        are taken care of.
        """
        project = tempfile.mkdtemp()
        with open(os.path.join(project, "MLProject"), 'w') as f:
            f.write("entry_points:\n  one:\n    parameters:\n      x: {type: float, default: 2}")
        # pass a parameter
        runner = ModelRunner()
        def my_model_1(dsl):
            assert dsl.get_parameter("x") == 1
        runner.set_parameter("x", 1)
        runner.run_method(my_model_1, path=project)
        # use default parameter
        runner = ModelRunner(entry_point="one")
        def my_model_2(dsl):
            assert dsl.get_parameter("x") == 2
        runner.run_method(my_model_2, path=project)
        # coercion of parameter to type declared in metadata
        runner = ModelRunner(entry_point="one")
        def my_model_3(dsl):
            assert dsl.get_parameter("x") == 1.5
        runner.set_parameter("x", "1.5")
        runner.run_method(my_model_3, path=project)
        # validation of parameter
        runner = ModelRunner(entry_point="one")
        def my_model_4(dsl):
            self.assertRaises(ValueError, lambda: dsl.get_parameter("x"))
        runner.set_parameter("x", "not numeric")
        runner.run_method(my_model_4, path=project)
        # clean up
        shutil.rmtree(project)

    def test_model_parameter_mapping(self):
        """
        Parameters are injected if they match the method's signature.  They are also coerced to the proper types.
        """
        runner = ModelRunner()
        def my_model(dsl, x: int, y: str):
            assert x == 1
            assert y == "2"
            assert dsl.get_parameter("z") == 3
        runner.set_parameter("x", 1).set_parameter("y", "2").set_parameter("z", 3)
        runner.run_method(my_model)
        runner.set_parameter("x", "1").set_parameter("y", 2).set_parameter("z", 3)
        runner.run_method(my_model)

    def test_model_parameter_json_parse_and_schema_check(self):
        """
        In addition to the very limited set of checks the MLProject file specifies, a full JSON schema can be given
        to validate more advanced structures, passed as JSON.
        """
        project = tempfile.mkdtemp()
        with open(os.path.join(project, "MLProject"), 'w') as f:
            f.write("entry_points:\n  one:\n    parameters:\n      x: {type: string, default: 2, schema: {type: object, properties: {a: {type: integer}}}}")
        runner = ModelRunner(entry_point="one")
        def my_model(dsl):
            assert dsl.get_parameter("x") == {"a": 4}
        runner.set_parameter("x", '{"a": 4}')
        runner.run_method(my_model, path=project)
        # clean up
        shutil.rmtree(project)

    def test_load_input_from_sql(self):
        """
        The caller can request that data be loaded from an SQL data source, instead of the usual CSV file.
        """
        runner = ModelRunner()
        log = []
        class Cursor(object):
            def __init__(self):
                self.description = ("a", None), ("b", None)
            def execute(self, sql):
                log.append(sql)
            def __iter__(self):
                return iter([(1, 2), (3, 4)])
        class Conn(object):
            def __init__(self, resource_name, username=None):
                assert username == "xyz"
                log.append(resource_name)
            def cursor(self):
                return Cursor()
            def close(self):
                log.append("close")
        def my_model_1(dsl):
            with mock.patch("dslibrary.transport.to_local.DSLibraryLocal.get_sql_connection", Conn):
                df = dsl.load_dataframe("x")
            assert list(df.columns) == ["a", "b"]
            assert list(df.a) == [1, 3]
            assert list(df.b) == [2, 4]
        runner.set_input("x", "sql:etc", sql_table="tbl1", username="xyz")
        runner.run_method(my_model_1)

    def test_log_metrics_to_csv(self):
        """
        Simple case of redirecting metrics to a CSV file.
        """
        project = tempfile.mkdtemp()
        runner = ModelRunner(uri="my_uri", run_id="run001", user="user3")
        runner.send_metrics_to("metrics.csv", format="csv")
        T = [1000]
        with mock.patch("time.time", lambda: T[0]):
            def my_model(dsl):
                dsl.log_metric("x", 123)
                T[0] += 1
                dsl.log_metric("y", 456)
            runner.run_method(my_model, path=project)
        metrics_fn = os.path.join(project, "metrics.csv")
        with open(metrics_fn, 'r') as f:
            data = f.read()
        assert data == 'uri,run_id,user,time,name,value,step\nmy_uri,run001,user3,1000,x,123,0\nmy_uri,run001,user3,1001,y,456,0\n'
        shutil.rmtree(project)

    def test_log_params_with_default_format(self):
        project = tempfile.mkdtemp()
        runner = ModelRunner(uri="my_uri", run_id="run001", user="user3")
        T = [1000]
        with mock.patch("time.time", lambda: T[0]):
            def my_model(dsl):
                dsl.log_param("a", 'one')
                T[0] += 1
                dsl.log_param("b", 222)
            runner.run_method(my_model, path=project)
        params_fn = os.path.join(project, PARAMS_ALIAS)
        with open(params_fn, 'r') as f:
            lines = f.read().strip().split("\n")
        assert json.loads(lines[0]) == {"uri": "my_uri", "run_id": "run001", "user": "user3", "time": 1000, "name": "a", "value": "one"}
        assert json.loads(lines[1]) == {"uri": "my_uri", "run_id": "run001", "user": "user3", "time": 1001, "name": "b", "value": 222}
        shutil.rmtree(project)

    def test_supply_and_capture_data(self):
        """
        You can pass in a dataframe!
        """
        def my_model(dsl):
            df = dsl.load_dataframe("input")
            dsl.log_metric("tot_before", df.x.sum())
            df.x += 1
            dsl.log_metric("tot_after", df.x.sum())
            dsl.write_resource("output", df)
        runner = ModelRunner()
        runner.set_input("input", data=pandas.DataFrame({"x": [1, 2, 3]}))
        # TODO we should probably have a global flag that captures everything, otherwise files will be created while debugging
        runner.set_output("output", capture=True)
        runner.send_metrics_to(capture=True)
        outputs = runner.run_method(my_model)
        self.assertEqual(list(outputs["output"].x), [2, 3, 4])
        metrics = outputs[METRICS_ALIAS]
        assert metrics == {'tot_before': 6, 'tot_after': 9}

    def tearDown(self) -> None:
        reset_env("test_run_method")



import unittest
import os
import tempfile
import shutil
import sys
import inspect
import pandas

from dslibrary.front import METRICS_ALIAS
from dslibrary.run_models import ModelRunner
from tests.t_utils import reset_env


class TestRunLocal(unittest.TestCase):
    """
    Execute models locally as subprocesses.
    """

    def test_run_local__process_csv(self):
        """
        Here we pass in a CSV and have it processed.
        """
        def my_model():
            import dslibrary
            dsl = dslibrary.instance()
            incr = dsl.get_parameter("incr")
            df = dsl.load_dataframe("input1")
            df.x += incr
            dsl.write_resource("output1", df)
        with RunLocalTestFramework(my_model) as env:
            # store input file
            env.write_file("i1", "x\n1\n2\n3")
            # NOTE: '~/i1' points to a file inside the model
            # NOTE: output2 is directed to a file outside the model by using an absolute path
            env.runner.set_input("input1", "~/i1", format="csv")
            env.runner.set_output("output1", os.path.join(env.output_folder, "o1"), format="csv")
            # metadata defines input & output, entry point, etc..
            interpreter = sys.executable
            with open(os.path.join(env.project, "MLProject"), 'w') as f:
                f.write("entry_points:\n  e1:\n    parameters:\n      incr: {type: float, default: 1}\n    command: \"" + interpreter + " model.py\"")
            env.run()
            # verify output was written
            self.assertEqual(env.read_file("o1"), 'x\n2.0\n3.0\n4.0\n')

    def test_infer_command_from_path(self):
        """
        ModelRunner() lets us execute a given source file without having to hook it up as an official entry point.
        So, we need to know how to execute various source files.
        """
        r = ModelRunner.infer_command_from_path("abc.py")
        assert r == [sys.executable, "abc.py"], r
        r = ModelRunner.infer_command_from_path("abc.r")
        assert r == ['RScript', 'abc.r'], r
        r = ModelRunner.infer_command_from_path("abc.ipynb")
        assert r == ['nbconvert', '--to', 'notebook', '--execute', '--inplace', 'abc.ipynb'], r

    def test_supply_and_capture_data(self):
        """
        You can pass in a dataframe from memory and it will send it using a temporary file.  And if you request to
        capture an output, or the metrics, they will be stored in temporary files, parsed, and returned in memory.
        """
        def my_model():
            import dslibrary as dsl
            df = dsl.load_dataframe("input")
            dsl.log_metric("tot_before", df.x.sum())
            df.x += 1
            dsl.log_metric("tot_after", df.x.sum())
            dsl.write_resource("output", df)
        with RunLocalTestFramework(my_model) as env:
            env.runner.set_input("input", data=pandas.DataFrame({"x": [1, 2, 3]}))
            env.runner.set_output("output", capture=True)
            env.runner.send_metrics_to(capture=True)
            outputs = env.run("model.py")
            assert outputs
            self.assertEqual(list(outputs["output"].columns), ["x"])
            self.assertEqual(list(outputs["output"].x), [2, 3, 4])
            metrics = outputs[METRICS_ALIAS]
            assert metrics == {'tot_before': 6, 'tot_after': 9}

    def tearDown(self) -> None:
        reset_env("test_run_local")


class RunLocalTestFramework(object):
    def __init__(self, function_with_code):
        self.function_with_code = function_with_code

    def __enter__(self):
        self.project = tempfile.mkdtemp()
        self.output_folder = tempfile.mkdtemp()
        self.runner = ModelRunner(uri="uri", entry_point="e1", project_root=self.project)
        model_code = inspect.getsource(self.function_with_code)
        # store model code
        self.write_file("model.py", model_code.strip() + f"\n\n{self.function_with_code.__name__}()")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        shutil.rmtree(self.project)
        shutil.rmtree(self.output_folder)

    def run(self, path: str=None):
        pkg_folder = os.path.join(os.path.dirname(__file__), "..")
        return self.runner.run_local(path, extra_env={"PYTHONPATH": pkg_folder})

    def write_file(self, path: str, content: str):
        with open(os.path.join(self.project, path), 'w') as f_w:
            f_w.write(content)

    def read_file(self, path: str) -> str:
        with open(os.path.join(self.output_folder, path), 'r') as f_r:
            return f_r.read()



"""
SNAPSHOT FROM https://innersource.accenture.com/projects/MSDOHEDCAIPDS/repos/model-management/browse/model-management-ds/src/main/mmlibraries/python/mmlibrary.py
"""

import glob
import json
import os
import sys
import uuid
import warnings

import requests
from bson.objectid import ObjectId
from gridfs import GridFSBucket
from pymongo import MongoClient

"""
Script that would parse the arguments passed from Model Manager
while executing any python job. Expected format of argument is as follows

python <scriptname.py> <job_id> secondary_id <?> model <model_file_absolute_path> input_file <input_data_file_abs_path> 
 -> KPI-START key1 95 key2 90 KPI-END PARAM-START param1 value1 param2 value2 PARAM-END

Date of creation: 8th March, 2017

Author: Model Management Dev Team
----------------------------------------------------------------------------------------------------------
Import libraries for downloading file from mongodb
"""

arguments = {}
kpi_dictionary = {}
param_dictionary = {}
input_file = {}
parsed_json_args = {}
env_config = {}
recipe_runtime = {}

__mm_arguments_parsed = False


def __custom_warning_format__(message, category, filename, lineno, file=None, line=None):
    sys.stdout.write(warnings.formatwarning(message, category, filename, lineno))


warnings.showwarning = __custom_warning_format__


def __evaluate_parse_arguments__():
    """
    Checks whether the <c>parse_arguments</c> function has already been called or not. If not then this function
    calls it making sure that the mmlibrary is properly initialized.
    """
    global __mm_arguments_parsed

    if __mm_arguments_parsed:
        return

    parse_arguments("")


def parseArguments(args=None):
    warnings.warn("Function has been renamed according to pep-0008 standard. "
                  "Please use parse_arguments(args=None) instead.")
    return parse_arguments(args)


def parse_arguments(args=None):
    """
    Parse arguments from sys.argv, for use with get_argument(key).
    This is not intended to use by users and deprecated from users point of view.
    :param args:   Currently ignored.
    """
    global __mm_arguments_parsed

    if __mm_arguments_parsed:
        return
    __mm_arguments_parsed = True

    global kpi_dictionary
    global param_dictionary
    global input_file
    global parsed_json_args
    global arguments
    global env_config
    global recipe_runtime

    param = __build_args__()

    parsed_json_args = json.loads(param)
    arguments = parsed_json_args["args"]
    kpi_dictionary = arguments["kpi"]
    param_dictionary = arguments["params"]
    input_file = arguments["input_file"]
    env_config = arguments["env_config"]
    recipe_runtime = parsed_json_args["runtime"]


# Added below for getting jobId, which is outside args json of job request.
def __get_job_id__():
    global parsed_json_args
    return parsed_json_args["job_id"]


def about():
    print("Model Manager header library v${python_version} (id: ${buildNumber.commitsha}, "
          "built: ${buildNumber.timestamp})")


def __get_json_argument__(key):
    global arguments
    try:
        return arguments[key]
    except KeyError:
        __raise_value_error__("System does not have a value for the argument key = " + key)


def __all_files_in__(dir_):
    all_files = None
    try:
        all_files = os.listdir(dir_)
    except os.error:
        __raise_value_error__("__get_all_files_in__: unable to retrieve files from dir: " + dir_)

    return all_files


def __score_model_binary_path__():
    score_base_dir = __get_env_config__("mm_score_mbin_fpath")
    model_resource_id = __get_json_argument__("model_resource_id")

    score_model_binary_cached_dir = os.path.join(score_base_dir, model_resource_id)
    # find all files in scoring_model_path directory
    all_files = __all_files_in__(score_model_binary_cached_dir)
    return os.path.join(score_model_binary_cached_dir, all_files[0])


def getModel():
    warnings.warn("Function has been renamed according to pep-0008 standard. "
                  "Please use get_model() instead.")
    return get_model()


def get_model():
    """
    Access data for the current model's model binary.
    :return: Path to model binary or ``None`` if it is not found.
    :except: ``ValueError`` if it is called from step recipe.
    """
    __evaluate_parse_arguments__()

    model = None
    model_path = None
    val = __get_json_argument__("model")
    is_not_model = __get_json_argument__("adhocStep")

    if is_not_model:
        __raise_value_error__("get_model() can only be called from model's recipe")

    if val is None:
        __raise_value_error__("System does not have a value for model")

    if __is_scoring_run__():
        model_path = __score_model_binary_path__()
    else:
        try:
            model_binary = __get_resource__(val)
            # check model is not null
            if model_binary is not None:
                with open(__temp_file_in_the_working_directory__(), "w+b") as model:
                    model.write(model_binary)

        except:
            __raise_value_error__("get_model: unable to retrieve model")

        if model is not None:
            model_path = model.name
            model.close()

    return model_path


def __get_resource__(file_id):
    """
    Return resourceContent as a stream of bytes.
    :param file_id: The gridfs file id.
    :return: returned bytes
    """
    return __stream_file_data_as_binary__(file_id).read()


def getArgument(key):
    warnings.warn("Function has been renamed according to pep-0008 standard. "
                  "Please use get_argument(key) instead.")
    return get_argument(key)


def get_argument(key):
    """
    Get the value of one of this model's supplied arguments.
    :param key: Name of argument.
    :return: Value of argument.
    :except: ``ValueError`` if the key is not contained by the dictionary.
    """
    global param_dictionary
    __evaluate_parse_arguments__()

    try:
        return param_dictionary[key]
    except KeyError:
        __raise_value_error__("System does not have a value for parameter key = " + key)


def newVersion(data, autoPublish=True):
    warnings.warn("Function has been renamed according to pep-0008 standard. "
                  "Please use new_version(data, auto_publish=True) instead.")
    return new_version(data, autoPublish)


def new_version(data, auto_publish=True):
    """
    Create new version for trained binary data, i.e. save model binary.
    :param data:  A readable stream with the content to save, i.e. a file-like object with a ``read()`` method.
    :param auto_publish: Whether the new trained version should be auto-published. Please note, this parameter is just an intent, if the current version is not published this version won't be published either even if the autopublish is TRUE
    :except: ``ValueError`` in the following cases:
    * if data is not file-like object or string or
    * if the master node endpoint is not set or
    * if the request sent to the master node was unsuccessful.
    """

    __evaluate_parse_arguments__()

    if not (data and (hasattr(data, "read") or isinstance(data, (str, bytes)))):
        __raise_value_error__("data is not provided or is of the wrong type")

    mm_endpoint = __get_master_node_endpoint__()
    if mm_endpoint is None:
        __raise_value_error__("model manager endpoint is not available")

    model_binary_name = "trainedmodel"
    mm_rest_url = mm_endpoint + "models/newtrainedversion"
    fs_bucket = __get_grid_fs_bucket__()
    fs_bucket.upload_from_stream(model_binary_name, data,
                                 metadata={
                                     "jobId": __get_job_id__(),
                                     "name": model_binary_name,
                                     "modelBinary": "true"
                                 })

    r = requests.post(mm_rest_url,
                      data=json.dumps({
                          "jobId": __get_job_id__(),
                          "secondary_id": __get_json_argument__('secondary_id'),
                          "autoPublish": auto_publish
                      }),
                      headers={'content-type': 'application/json'})

    if r.status_code != 200 or r.json().get("model_id") is None:
        __raise_value_error__("Error while creating new version of the training model")


def save_kpi(kpi_name, kpi_value, model_name=None, model_version_id="latest"):
    """
    Save KPI result value - for KPI name, model and model version.
    :param kpi_name the kpi name
    :param kpi_value the kpi result value to store
    :param model_name the name of the model
    :param model_version_id the version id
    :except ``ValueError`` in the following case:
    * if the request sent to the master node wasn't successful.
    """

    __evaluate_parse_arguments__()

    is_step = __get_json_argument__("adhocStep")

    ids = list(__get_json_argument__('secondary_id').split("_"))
    run_id = ids[1]
    model_id = ""
    version_id = model_version_id

    if is_step:
        if model_name is None or version_id is None:
            __raise_value_error__("Save Kpi: method save_kpi(kpi_name, kpi_value) "
                                  "should be used within model only. "
                                  "Use method save_kpi(kpi_name, kpi_value, model_name, "
                                  "model_version_id) within step.")
    else:  # model
        if model_name is not None or version_id != "latest":
            __raise_value_error__("Save Kpi: method save_kpi(kpi_name, kpi_value, "
                                  "model_name, model_version_id) should be used within step only. "
                                  "Use method save_kpi(kpi_name, kpi_value) "
                                  "within a model.")
        else:
            model_id = ids[2]  # modelId
            version_id = ids[3]  # versionId

    mm_endpoint = __get_master_node_endpoint__()
    mm_rest_url = mm_endpoint + "kpiresults"

    r = requests.post(mm_rest_url,
                      data=json.dumps({"kpi_result_value": kpi_value,
                                       "kpi_name": kpi_name,
                                       "model_id": model_id,
                                       "model_name": model_name,
                                       "model_version_id": version_id,
                                       "project": get_project(),
                                       "run_id": run_id}),
                      headers={'content-type': 'application/json'})

    if r.status_code != 200:
        json_response = r.json()
        if len(json_response) > 0:
            __raise_value_error__("Save Kpi: " + json_response[0]["ValidationError"])
        __raise_value_error__("Save Kpi: Unable to save KPI result")


def saveKpiResult(kpiName, referenceDate, kpiResultValue, modelName=None, versionId=None):
    """
    Save KPI result value - for KPI name, reference date, model and model version.
    :param kpiName the kpi name
    :param referenceDate the reference date by the ``getKpiResult`` function can retrieve result.
    IGNORED since 5.2 release, the creation date of the KpiResult will be used instead of!
    :param kpiResultValue the kpi result value to store
    :param modelName the name of the model
    :param versionId the version id
    :except ``ValueError`` in the following cases:
    * if the reference date is given and it is not in the expected format or
    * if the request sent to the master node wasn't successful.
    """

    warnings.warn("Deprecated function, it may be removed in future releases.")

    __evaluate_parse_arguments__()

    is_step = __get_json_argument__("adhocStep")

    # extract additional parameters
    ids = list(__get_json_argument__('secondary_id').split("_"))
    run_id = ids[1]  # runId
    model_id = ""
    model_version_id = versionId
    if is_step:
        if modelName is None or versionId is None:
            __raise_value_error__("saveKpiResult: method saveKpiResult(kpiName, referenceDate, kpiResultValue) "
                                  "should be used within model only. "
                                  "Use method saveKpiResult(kpiName, referenceDate, kpiResultValue, modelName, "
                                  "versionId) within step.")
    else:  # model
        if modelName is not None or versionId is not None:
            __raise_value_error__("saveKpiResult: method saveKpiResult(kpiName, referenceDate, kpiResultValue, "
                                  "modelName, versionId) should be used within step only. "
                                  "Use method saveKpiResult(kpiName, referenceDate, kpiResultValue) "
                                  "within a model.")
        else:
            model_id = ids[2]  # modelId
            model_version_id = ids[3]  # versionId

    mm_endpoint = __get_master_node_endpoint__()
    mm_rest_url = mm_endpoint + "kpiresults"
    r = requests.post(mm_rest_url,
                      data=json.dumps({"kpi_result_value": kpiResultValue,
                                       "kpi_name": kpiName,
                                       "model_id": model_id,
                                       "model_name": modelName,
                                       "model_version_id": model_version_id,
                                       "project": get_project(),
                                       "run_id": run_id}),
                      headers={'content-type': 'application/json'})

    if r.status_code != 200:
        json_response = r.json()
        if len(json_response) > 0:
            __raise_value_error__("saveKpiResult: " + json_response[0]["ValidationError"])
        __raise_value_error__("saveKpiResult: Unable to save KPI result")


def get_last_kpi(kpi_name, model_name, model_version_id="all"):
    """
    Get the value of the latest KPI.
    :param kpi_name The kpi name criteria.
    :param model_name The model name criteria.
    :param model_version_id The version id criteria. It can be "all", "latest" or numeric.
    :return With a dictionary containing the kpi result and timestamp retrieved by the criteria or
            ``None`` if not found any.
    :except ``ValueError`` if the request sent to the master node was unsuccessful.
    """
    __evaluate_parse_arguments__()

    mm_endpoint = __get_master_node_endpoint__()
    mm_rest_url = mm_endpoint + "kpiresults/last"
    project = get_project()
    response = requests.get(mm_rest_url,
                            params={
                                'model_name': model_name,
                                'version_id': model_version_id,
                                'kpi_name': kpi_name,
                                'project': project
                            })

    json_response = response.json()
    if response.status_code != 200:
        if len(json_response) > 0:
            __raise_value_error__("get_last_kpi: " + json_response[0]["ValidationError"])
        __raise_value_error__("get_last_kpi: Error while get KPI result")

    if (json_response is None or
            json_response.get("kpi_result_value") is None or
            json_response.get("created") is None):
        return None

    return {
        "value": json_response.get("kpi_result_value"),
        "timestamp": json_response.get("created")
    }


def getKpiResult(kpiName, modelName, versionId, referenceDate):
    """
    Get the value of a KPI.
    :param kpiName The kpi name criteria.
    :param modelName The model name criteria.
    :param versionId The version id criteria.
    :param referenceDate The reference date criteria. IGNORED since 5.2 release!
    :return With the kpi result retrieved by the criteria or ``None`` if not found any.
    :except ``ValueError`` if the ``referenceDate`` is not given in the expected format or
    the request sent to the master node was unsuccessful.
    """

    warnings.warn("Deprecated function, it may be removed in future releases. The <referenceDate> field is ignored.")

    return get_last_kpi(kpiName, modelName, versionId)


def getBinaryFromResource(resourceName):
    warnings.warn("Function has been renamed according to pep-0008 standard. "
                  "Please use get_binary_from_resource(resource_name) instead.")
    return get_binary_from_resource(resourceName)


def get_binary_from_resource(resource_name):
    """
    Download the entire content of the given resource from mongodb.
    :param resource_name:  Name of resource to download.
    :returns:  Binary content.
    :except: ``ValueError`` if the resource requested by name doesn't exist or its content couldn't be retrieved.
    """
    # input_file is Json array
    __evaluate_parse_arguments__()

    file_id = __get_resource_file_id__(resource_name)
    # Download the file content to a local file and then pass the file name to the caller
    content = __get_resource__(file_id)
    if content is None:
        __raise_value_error__("get_binary_from_resource: unable to retrieve data from resource: <" + resource_name + ">")
    return content


def saveBinaryToResource(resourceName, data):
    warnings.warn("Function has been renamed according to pep-0008 standard. "
                  "Please use save_binary_to_resource(resource_name, data) instead.")
    return save_binary_to_resource(resourceName, data)


def save_binary_to_resource(resource_name, data):
    """
    Upload binary resource file to mongodb.
    :param resource_name:  Name of resource.
    :param data: A file-like object with a ``read()`` method or ``string``.
    :except: ``ValueError`` in the following cases:
    * if the ``resourceName`` parameter is not provided or ``None`` or
    * if the ``data`` parameter is neither file-like object nor ``string``.
    """
    __evaluate_parse_arguments__()

    if not resource_name:
        __raise_value_error__("resource name is not provided")

    if not data or not (hasattr(data, "read") or isinstance(data, (str, bytes))):
        __raise_value_error__("data is not provided or is of the wrong type")

    fs_bucket = __get_grid_fs_bucket__()

    # output_file is Json array
    fs_bucket.upload_from_stream(resource_name,
                                 data.encode('utf-8') if isinstance(data, str) else data,
                                 metadata={'jobId': __get_job_id__(),
                                           'name': resource_name})


def saveTemporaryData(run_data):
    warnings.warn("Function has been renamed according to pep-0008 standard. "
                  "Please use save_temporary_data(run_data) instead.")
    return save_temporary_data(run_data)


def save_temporary_data(run_data):
    """
    Uploads transient resource file to mongodb for given run.
    :param run_data: A file-like object with a ``read()`` method or ``string``.
    :type run_data: str or bytes or file-like object using either in binary or text mode
    :except: ``ValueError`` in the following cases:
    * if the ``run_data`` parameter is neither file-like object nor ``string``,
    * if the ``transient_resource_id`` is ``None``,
    * if the ``run_data`` couldn't be uploaded through the Resource API.
    """
    __evaluate_parse_arguments__()

    if not run_data or not (hasattr(run_data, "read") or isinstance(run_data, (str, bytes))):
        __raise_value_error__("run_data is not provided or is of the wrong type")

    transient_resource_id = __get_transient_resource_id__()
    if transient_resource_id is None:
        __raise_value_error__("The transient_resource_id is missing.")

    save_transient_resource_url = \
        __get_master_node_endpoint__() + "resources/transient_resource/" + transient_resource_id

    if isinstance(run_data, (str, bytes)):
        file_data = ("transient_resource_data_file_name",
                     run_data.encode() if isinstance(run_data, str) else run_data)
    else:
        file_data = run_data

    response = requests.post(url=save_transient_resource_url, files={'fileData': file_data})

    if response.status_code == 400:
        __raise_value_error__("save_temporary_data: " + response.json()[0]["ValidationError"])
    elif response.status_code != 200:
        __raise_value_error__("save_temporary_data: transient resource data couldn't be uploaded.")


def __get_response_data__(response):
    content = b''
    for chunk in response.iter_content(chunk_size=128):
        content += chunk

    return content


def getTemporaryData():
    warnings.warn("Function has been renamed according to pep-0008 standard. "
                  "Please use get_temporary_data() instead.")
    return get_temporary_data()


def get_temporary_data():
    """
    Downloads transient resource file from mongodb for given run.
    :returns The content of the downloaded temporary data.
    :rtype bytes
    """
    __evaluate_parse_arguments__()

    transient_resource_id = __get_transient_resource_id__()
    if transient_resource_id is None:
        __raise_value_error__("The transient_resource_id is missing.")

    download_transient_resource_url = \
        __get_master_node_endpoint__() + "resources/" + transient_resource_id + "/download"

    response = requests.get(download_transient_resource_url, stream=True)

    if response.status_code != 200:
        json_response = response.json()
        if len(json_response) > 0:
            __raise_value_error__("get_temporary_data: " + json_response[0]["ValidationError"])
        __raise_value_error__("get_temporary_data: transient resource data couldn't be downloaded.")

    return __get_response_data__(response)


# ---- Private methods --------------
def __get_grid_fs_bucket__():
    db = __get_model_manager_db__()
    return GridFSBucket(db, bucket_name="mm.fs")


def __get_resource_file_id__(resource_name):
    """
    Determine the resource file id
    """
    global input_file
    # input_file is Json array
    file_id = None
    for file in input_file:
        if file['name'] == resource_name:
            file_id = file['value']
    if file_id is None:
        __raise_value_error__("System does not have a value for input file key = " + resource_name +
                              ", please check the uploaded resource(s) name(s)")
    return file_id


def __stream_file_data_as_binary__(file_id):
    fs_bucket = __get_grid_fs_bucket__()
    return fs_bucket.open_download_stream(ObjectId(file_id))


def __build_args__():
    arg = ""
    for index in range(len(sys.argv) - 1):
        arg = arg + " " + sys.argv[index + 1]
    return arg


class MmLibraryValidationWarning(UserWarning):
    pass


def evaluationResult(message):
    warnings.warn("Function has been renamed according to pep-0008 standard. "
                  "Please use evaluation_result(message) instead.")
    return evaluation_result(message)


def evaluation_result(message):
    """
    It is used to indicate evaluation failure scenarios.
    :param message: optional. The message to use in the printed warning.
    """
    __evaluate_parse_arguments__()

    if message is not None:
        warnings.warn("mmlibrary evaluation error: " + message, MmLibraryValidationWarning, stacklevel=2)
    else:
        warnings.warn("mmlibrary evaluation error: ", MmLibraryValidationWarning, stacklevel=2)


def getDBConnection(jdbc_connection_name, create_connection=None):
    warnings.warn("Function has been renamed according to pep-0008 standard. "
                  "Please use get_db_connection(jdbc_connection_name, create_connection=None) instead.")
    return get_db_connection(jdbc_connection_name, create_connection)


def get_db_connection(jdbc_connection_name, create_connection=None):
    """
    Get a connection to a database.
    :param jdbc_connection_name:   Name associated with connection details.
    :param create_connection: An optional callable object, so it can be function or lambda expression, too. The callable
    object must have at least 3 parameters if it is provided. The first where the connection url, the second where the
    user name and the third where the password will be passed.
    If this callable object is provided then its return value will be used as a connection to the database. Otherwise
    the get_db_connection function will create a connection to it.
    If a connection name provided as first parameter which refers to a JDBC connection then this parameter will be
    ignored.
    :return:   A standard python database connection object. The user's responsibility is to close the database connection.
    :except: * ``ValueError`` either if the db configuration cannot be retrieved or
    connection attempt to the database was unsuccessful or
    the database connection name refers to a non-JDBC connection where the connection type is ``postgresql`` and the
    ``psycopg2`` library is not available/installed.
    * ``KeyError`` if the user refers to DB connection details which doesn't exist in the system.
    """

    __evaluate_parse_arguments__()

    db_connection_details = __get_db_connection_details__(jdbc_connection_name, get_project())

    if db_connection_details["custom"] and create_connection is None:
        print("Detected custom connection without connection_creation function")
        __raise_value_error__("A custom db connection requires the use of the connection_creation function")

    if __connection_url_type_is_jdbc__(db_connection_details["url"]):
        print("Detected JDBC connection -- checking requirements")
        return __jaydebeapi_connection__(db_connection_details, jdbc_connection_name)

    if create_connection is not None:
        print("Detected custom connection creation function -- checking requirements")
        return __custom_db_connection__(create_connection, db_connection_details)

    if db_connection_details["type"] == "postgresql":
        print("Detected native PostgreSQL connection -- checking requirements")
        if __is_psycopg2_importable__():
            return __psycopg2_connection__(db_connection_details, jdbc_connection_name)
        else:
            __raise_mmlibrary_error__("MMLIB_ERR_POSTGRESQL_DRIVER_MISSING",
                                      "psycopg2 library is not installed in the used Python environment.",
                                      "Install the psycopg2 library. E.g. pip install psycopg2-binary")

    if db_connection_details["type"] == "teradata":
        print("Detected native Teradata Database connection -- checking requirements")
        if __is_teradata_sql_importable__():
            return __teradata_connection__(db_connection_details, jdbc_connection_name)
        else:
            __raise_mmlibrary_error__("MMLIB_ERR_TERADATA_DRIVER_MISSING",
                                      "teradatasql library is not installed in the used Python environment.",
                                      "Install the teradatasql library. E.g. pip install teradatasql")


def __get_runtime__():
    return recipe_runtime


def __get_db_connection_details__(db_connection_name, project):
    db_data = __get_db_data__()
    db_connection_details = None
    try:
        db_connection_details = db_data[project][db_connection_name]
    except KeyError:
        __raise_value_error__("System does not have a DB connection with the given connection name = " +
                              db_connection_name + " and project = " + project)

    return db_connection_details


def __get_db_config_details__(db_type):
    db = __get_model_manager_db__()
    runtime = __get_python_runtime__()
    db_details = db["mm.dbconfigdetails"].find({'runtime': runtime})
    if db_details is None:
        __raise_value_error__("No connection found for runtime = " + runtime)

    jar_list = list()
    data_base_details = [None, None]
    for db_detail in db_details:
        if db_detail["drv_jar_location"] is not None:
            jar_list.extend(glob.glob(db_detail["drv_jar_location"] + "/*.jar"))
        if db_type == db_detail["type"]:
            data_base_details[0] = db_detail["driver_class"]

    data_base_details[1] = ":".join(jar_list)
    return data_base_details


def __get_model_manager_db__():
    mongo_db_uri = __get_mongo_db_uri__()
    client = MongoClient(mongo_db_uri)
    return client.get_default_database()


def __get_db_data__():
    db_data = __get_env_config__("db_data")

    if not isinstance(db_data, dict):
        db_data = json.loads(db_data)

    return db_data


def __get_mongo_db_uri__():
    return __get_env_config__("mongoURI")


def __get_env_config__(key):
    global env_config
    try:
        return env_config[key]
    except KeyError:
        __raise_value_error__("System does not have a value for parameter key = " + key)


def __get_python_runtime__():
    """
    Determine python runtime
    :return: 'container' for containerized environment, 'python' for 'python 3'
    """
    runtime = __get_runtime__()
    if runtime == "container":
        return 'container'
    return 'python'


def __custom_db_connection__(create_connection, db_connection_details):
    print("Creating custom connection.")

    import inspect
    if not inspect.isfunction(create_connection):
        __raise_value_error__("The create_connection parameter must be a callable object. But it is " +
                              str(type(create_connection)) + ".")

    parameter_number = len(inspect.signature(create_connection).parameters)
    if parameter_number < 3:
        __raise_value_error__("The create_connection parameter must be a callable object whose has at least 3 " +
                              "parameters but it has only " + str(parameter_number) + ".")

    connection = create_connection(db_connection_details["url"],
                                   db_connection_details["user_name"],
                                   db_connection_details["password"])
    if connection is None:
        print("Warning: the created custom connection is seems to be not a valid one.")
    else:
        print("...done.")

    return connection


def __connection_url_type_is_jdbc__(connection_url):
    return connection_url.startswith("jdbc:")


def __is_psycopg2_importable__():
    try:
        import psycopg2
        return True
    except ImportError:
        return False


def __psycopg2_connection__(db_connection_details, db_connection_name):
    print("Creating native connection to PostgreSQL Database.")

    import psycopg2
    try:
        connection = psycopg2.connect(db_connection_details["url"],
                                      user=db_connection_details["user_name"],
                                      password=db_connection_details["password"])
        print("...done.")
        return connection
    except psycopg2.Error as error:
        __raise_value_error__("Cannot connect to " + db_connection_name + ". " + str(error))


def __is_teradata_sql_importable__():
    try:
        import teradatasql
        return True
    except ImportError:
        return False


def __teradata_connection__(db_connection_details, db_connection_name):
    print("Creating native connection to Teradata Database.")

    url_components = __connection_url_components__(db_connection_details["url"])
    json_connection_string = __json_connection_string__(db_connection_details, url_components)

    import teradatasql
    try:
        connection = teradatasql.connect(json_connection_string)
        print("...done.")
        return connection
    except teradatasql.Error as error:
        __raise_value_error__("Cannot connect to " + db_connection_name + ". " + str(error))


def __connection_url_components__(url):
    import urllib.parse
    url_components = urllib.parse.urlsplit(url)

    query_string = url_components.query.split("&") if len(url_components.query) > 0 else []
    query = list(map(lambda query_element: query_element.split("="), query_string))

    return {
        "host": url_components.hostname,
        "dbs_port": str(url_components.port),
        "database": url_components.path.lstrip("/") if url_components.path.startswith("/") else url_components.path,
        "query": query
    }


def __json_connection_string__(db_connection_details, url_components):
    joined_key_value_pairs = __query_part__(url_components["query"])

    return ('{"host": "' + url_components["host"] + '",' +
            ' "dbs_port": "' + url_components["dbs_port"] + '",' +
            ' "database": "' + url_components["database"] + '",' +
            ' "user": "' + db_connection_details["user_name"] + '",' +
            ' "password": "' + db_connection_details["password"] + '"' +
            joined_key_value_pairs +
            '}')


def __query_part__(query_list):
    if len(query_list) == 0:
        return ''

    query_filter = filter(lambda list_element: len(list_element) == 2 and len(list_element[1]) > 0,
                          query_list)
    key_value_pairs = map(lambda query_key_value: '"' + query_key_value[0] + '": "' + query_key_value[1] + '"',
                          query_filter)

    from functools import reduce
    return reduce(lambda first, second: first + ", " + second, key_value_pairs, '')


def __jaydebeapi_connection__(db_connection_details, db_connection_name):
    print("Creating JDBC connection.")

    db_details = __get_db_config_details__(db_connection_details["type"])
    if db_details is None:
        __raise_value_error__("Cannot get Db config details.")

    import jaydebeapi
    try:
        connection = jaydebeapi.connect(jclassname=db_details[0],
                                        url=db_connection_details["url"],
                                        driver_args=[db_connection_details["user_name"],
                                                     db_connection_details["password"]],
                                        jars=db_details[1])
        print("...done.")
        return connection
    except:
        __raise_value_error__("Cannot connect to " + db_connection_name)


def getProject():
    warnings.warn("Function has been renamed according to pep-0008 standard. "
                  "Please use get_project() instead.")
    return get_project()


def get_project():
    """
    Get the ID of the associated project.
    :except: ``ValueError`` if the project name can not be found.
    """
    __evaluate_parse_arguments__()

    project = __get_json_argument__("project")
    if project is None:
        __raise_value_error__("System does not know the project")
    return project


def __get_master_node_endpoint__():
    """
    Gives back the master node endpoint.
    :except ``ValueError`` if the master node endpoint value is not stored in the ``env_config`` dictionary.
    """
    return __get_env_config__("mm_master_node_endpoint")


def returnScore(score):
    warnings.warn("Function has been renamed according to pep-0008 standard. "
                  "Please use return_score(score) instead.")
    return return_score(score)


def return_score(score):
    """
    Returns the score result.
    :param score: the value to store in the output file.
    :except ``ValueError`` if this function called from not a scoring recipe.
    """
    __evaluate_parse_arguments__()

    if not __is_scoring_run__():
        __raise_value_error__("return_score: function can only be called inside scoring recipe")

    score_json = __get_score_json__(score)

    with open(__temp_file_in_the_working_directory__(file_name="scoreresult.txt"), "w+") as f:
        f.write(score_json)
        f.close()

    return score_json


def __get_transient_resource_id__():
    return __get_json_argument__("transient_resource_id")


def __get_score_json__(score):
    score_type = 'object'
    if __is_primitive__(score):
        score_type = 'primitive'
    score_info = {'type': score_type, 'value': score}
    return json.dumps(score_info)


def __is_primitive__(result_type):
    if isinstance(result_type, int):
        return True
    if isinstance(result_type, str):
        return True
    if isinstance(result_type, bool):
        return True
    if isinstance(result_type, float):
        return True
    if result_type is None:
        return True
    return False


def getrunid():
    warnings.warn("Function has been renamed according to pep-0008 standard. "
                  "Please use get_run_id() instead.")
    return get_run_id()


def get_run_id():
    """
    Returns the run id associated with the particular step/model run.
    :except ``ValueError`` if it is called from a scoring recipe or ``secondary_id`` is not stored.
    """
    __evaluate_parse_arguments__()

    if __is_scoring_run__():
        __raise_value_error__("get_run_id: function can not be called inside scoring recipe")

    # extract additional parameters
    ids = list(__get_json_argument__('secondary_id').split("_"))
    run_id = ids[1]  # runId
    return run_id


def __is_scoring_run__():
    """
    Determine whether the current run is a scoring run.
    :return: True if it is scoring otherwise false.
    """
    step_number = __get_json_argument__("stepNumber")
    entity_number = __get_json_argument__("entityNumber")

    return step_number is None and entity_number is None


def __get_script_tmp_location__():
    """
    Returning the script location
    :return: the directory name of the script
    """
    return os.path.dirname(os.path.abspath(sys.argv[0]))


def __temp_file_in_the_working_directory__(file_name=None):
    """
    Creating a temporary file that will be cleaned up upon a process exit
    :return: a temporary file
    """
    temp_file_name = file_name
    if temp_file_name is None:
        temp_file_name = str(uuid.uuid4())
    return os.path.join(__get_script_tmp_location__(), temp_file_name)


def __raise_value_error__(error_string):
    e = ValueError("mmlibrary: " + error_string)
    print(e)
    raise e


def __raise_mmlibrary_error__(code, description, advice):
    raise ValueError(__syntax__(
        [
            "__MMLIBRARY_ERROR_BEGINS__",
            json.dumps({
                "code": code,
                "description": description,
                "advice": advice
            }),
            "__MMLIBRARY_ERROR_ENDS__"
        ]
    ))


def __syntax__(iterable):
    return "\n".join(iterable)


if __name__ == '__main__':
    print("Executing header/installer python script")


import unittest

from dslibrary import parse_verb_and_args, instantiate_class_and_args, DSLibrary
from dslibrary.transport.to_local import DSLibraryLocal
from dslibrary.utils.instantiation import import_named_object, valueize


class TestInstantiation(unittest.TestCase):

    def test_valueize(self):
        assert valueize("1") == 1
        assert valueize("1.5") == 1.5
        assert valueize("1x") == "1x"

    def test_parse_verb_and_args(self):
        verb, args, kwargs = parse_verb_and_args("A:v1:x=1:y=2:z=z")
        assert verb == "A"
        assert args == ("v1",)
        assert kwargs == {"x": 1, "y": 2, "z": "z"}

    def test_instantiate_class_and_args(self):
        dsl = instantiate_class_and_args("dslibrary.transport.to_local.DSLibraryLocal:ROOT", DSLibrary)
        assert isinstance(dsl, DSLibraryLocal)
        assert dsl._root == "ROOT"

    def test_instantiate_class_and_args__wrong_type(self):
        self.assertRaises(ValueError, lambda: instantiate_class_and_args("pathlib.Path:abc", DSLibrary))

    def test_import_named_object(self):
        loads = import_named_object("json.loads")
        assert loads("{}") == {}




"""
Interfaces to remote engines, i.e. filesystems, SQL & NoSQL databases, etc..
"""
import pathlib
import typing
import io


class FileSystem(object):
    """
    A very simplified version of fsspec.AbstractFileSystem
    """
    def mkdir(self, path, create_parents=True, **kwargs):
        pass

    def rmdir(self, path):
        pass  # not necessary to implement, may not have directories

    def ls(self, path="", detail=True, **kwargs):
        """
        List files.  Returns a list of entries with these fields:
          'name' - full path to the entry (without protocol)
          'size' - size of the entry, in bytes, if it can be determined
          'type' - type of entry, "file", "directory" or other
        """

    def stat(self, path, **kwargs):
        # dict with keys: name (full path in the FS), size (in bytes), type (file,
        # directory, or something else) and other FS-specific keys.
        pass

    def exists(self, path, **kwargs):
        """Is there a file at the given path"""
        try:
            self.stat(path, **kwargs)
            return True
        except FileNotFoundError:
            return False

    def rm(self, path, recursive=False, maxdepth=None):
        pass

    def open(self, path: str, mode: str="rb", **kwargs):
        """
        Open a file for read, write or append.

        :param path: File to open, or URL.
        :param mode: r, rb, w, wb, a, ab
        :param kwargs: Customizable arguments...
            block_size
            cache_options
        :return:    A file-like object with an __exit__ method.
        """


class WriteNotAllowed(Exception):
    pass


class FileSystemLocal(FileSystem):
    """
    A chroot'ed set of local files.
    """
    def __init__(self, root: str):
        self.root = pathlib.Path(root)

    def _loc(self, path: str):
        out = self.root / path.strip("/")
        if ".." in out.parts:
            raise ValueError(".. not allowed")
        return out

    def mkdir(self, path, create_parents=True, **kwargs):
        self._loc(path).mkdir(parents=create_parents, mode=kwargs.get('mode'))

    def rmdir(self, path):
        self._loc(path).rmdir()

    def ls(self, path=".", detail=True, **kwargs):
        out = []
        for f in self._loc(path).glob("*"):
            out.append(self._file_info(f))
        return out

    @staticmethod
    def _file_info(f):
        return {
            "name": f.name,
            "size": f.stat().st_size,
            "type": "directory" if f.is_dir() else "symlink" if f.is_symlink() else "file"
        }

    def stat(self, path, **kwargs):
        return self._file_info(self._loc(path))

    def rm(self, path, recursive=False, maxdepth=None):
        # TODO support recursive directory removal
        self._loc(path).unlink(missing_ok=True)

    def open(self, path, mode="rb", **kwargs):
        return self._loc(path).open(mode=mode)


class FileSystemReadOnly(FileSystem):
    """
    Block write access.
    """
    def __init__(self, target: FileSystem):
        self._target = target

    def mkdir(self, path, create_parents=True, **kwargs):
        raise WriteNotAllowed()

    def rmdir(self, path):
        raise WriteNotAllowed()

    def ls(self, *args, **kwargs):
        return self._target.ls(*args, **kwargs)

    def stat(self, *args, **kwargs):
        return self._target.stat(*args, **kwargs)

    def rm(self, path, recursive=False, maxdepth=None):
        raise WriteNotAllowed()

    def open(self, path, mode="rb", **kwargs):
        if "w" in mode or "a" in mode:
            raise WriteNotAllowed()
        return self._target.open(path, mode=mode, **kwargs)


class SqlDatabase(object):
    """
    The methods expected of a database connection.  These are the methods one usually expects to exist in the connection
    object returned by a DBI-compliant driver.
    """
    def cursor(self):
        """ return a cursor """
    def close(self):
        """ clean up """
    def commit(self):
        """ commit changes (placeholder) """
    def rollback(self):
        """ roll back changes (placeholder) """


class NoSqlDatabase(object):
    """
    The methods expected for a NoSQL database.
    """
    def query(self, collection: str, query: dict=None, limit: int=None, **kwargs) -> typing.Iterable[dict]:
        """
        Scan for documents, or retrieve a specific document.
        :param collection:   Name of collection.
        :param query:        JSON-style query with named fields to exactly match.  Follows an implementation-dependent
                             subset of the MongoDB conventions.  Example: {"x": {"$lt": 100}}
        :param limit:        Maximum number of records.
        :return:    An iteration of matched documents.
        """

    def insert(self, collection: str, doc: dict, **kwargs):
        """
        Add a document to a collection.  Note that some systems have restrictions on field names.  For instance,
        MongoDB does not allow fields to start with '$'.

        :returns:  ID of new row.
        """

    def update(self, collection: str, filter: dict, changes: dict, upsert: bool=False, **kwargs):
        """
        Update documents in a collection.

        :param collection:  Name of collection we're working on.
        :param filter:      Which documents to update.
        :param changes:     A {} with named changes to make.  An implementation-dependent subset of MongoDB conventions
                            will be supported.  For instance, {"$inc": {"x": 3}}.
        """

    def delete(self, collection: str, filter: dict, **kwargs):
        """
        Delete documents from a collection.

        :param collection:  Name of collection we're working on.
        :param filter:      Which documents to delete.
        """


class NoSQLReadOnly(NoSqlDatabase):
    """
    Block write access.
    """
    def __init__(self, target: NoSqlDatabase):
        self._target = target

    def query(self, *args, **kwargs):
        return self._target.query(*args, **kwargs)

    def insert(self, collection: str, doc: dict, **kwargs):
        raise WriteNotAllowed()

    def update(self, collection: str, filter: dict, changes: dict, upsert: bool=False, **kwargs):
        raise WriteNotAllowed()

    def delete(self, collection: str, filter: dict, **kwargs):
        raise WriteNotAllowed()


class S3Connection(FileSystem):
    """
    Implementation of s3 connection.
    """
    def __init__(self, bucket: str, access_key: str, secret_key: str, read_only: bool=False):
        import boto3
        client = boto3.client('s3', aws_access_key_id=access_key, aws_secret_access_key=secret_key)
        self._client = client
        self._bucket = bucket
        self._read_only = read_only

    def open(self, path: str, mode: str="rb", **kwargs):
        from dslibrary.utils.file_utils import write_stream_with_read_on_close
        if "r" in mode:
            resp = self._client.get_object(Bucket=self._bucket, Key=path)
            stream = resp["Body"]
            if mode == "r":
                stream = io.TextIOWrapper(stream, 'utf-8')
            return GenericStreamWrapper(stream, writable=False)
        if self._read_only:
            raise PermissionError("this data source is read-only")
        if "w" in mode:
            # TODO this is not a 'streaming upload' -- it writes to a local file, then uploads that file when it is closed
            return write_stream_with_read_on_close(mode, 'rb', on_close=lambda f_r: self._client.upload_fileobj(f_r, Bucket=self._bucket, Key=path))
        if "a" in mode:
            raise NotImplementedError("append not supported for s3")

    def ls(self, path=None, detail=True, **kwargs):
        more = {}
        if path:
            more["Prefix"] = path
        resp = self._client.list_objects(Bucket=self._bucket, Delimiter="/", **more)
        contents = resp.get("Contents", [])
        prefixes = resp.get("CommonPrefixes", [])
        for obj in contents:
            yield {
                "name": obj["Key"],
                "size": obj["Size"],
                "modified": obj["LastModified"].timestamp(),
                "type": "file"
            }
        for obj in prefixes:
            yield {
                "name": obj["Prefix"],
                "type": "directory"
            }

    def stat(self, path, **kwargs):
        raise NotImplementedError()

    def exists(self, path, **kwargs):
        raise NotImplementedError()

    def rm(self, path, **kwargs):
        raise NotImplementedError()


class GenericStreamWrapper(io.RawIOBase):
    def __init__(self, stream, readable=None, writable=None):
        self._target = stream
        self._readable = readable if readable is not None else True if not hasattr(stream, "readable") else stream.readable()
        self._writable = writable if writable is not None else False if not hasattr(stream, "writable") else stream.writable()
        self._seekable = False if not hasattr(stream, "seekable") else stream.seekable()
        self._closed = False

    def readable(self) -> bool:
        return self._readable

    def writable(self) -> bool:
        return self._writable

    def seekable(self) -> bool:
        return self._seekable

    @property
    def closed(self):
        if not hasattr(self._target, "closed"):
            return self._closed
        return self._target.closed

    def close(self):
        self._closed = True
        if not hasattr(self._target, "close"):
            return
        self._target.close()

    def read(self, __size: int=None):
        return self._target.read(__size)

    def write(self, __b):
        if not hasattr(self._target, "write"):
            raise NotImplementedError()
        return self._target.write(__b)

    def seek(self, __offset: int, __whence: int=None) -> int:
        if not hasattr(self._target, "seek"):
            raise NotImplementedError()
        return self._target.seek(__offset, __whence)

    def flush(self) -> None:
        if not hasattr(self._target, "flush"):
            return
        self._target.flush()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



"""
Connections to NoSQL databases.
"""
import typing
from dslibrary.engine_intf import NoSqlDatabase, NoSQLReadOnly


def connect_to_nosql(uri: str, library: str=None, for_write: bool=False, **kwargs) -> NoSqlDatabase:
    """
    Connect to a NoSQL database based on a URI.
    """
    protocol = uri.split("://")[0]
    if protocol == "monogodb":
        try:
            import pymongo
        except ImportError:
            raise ImportWarning(f"pymongo library not available, can't open database: {uri}")
        db = pymongo.MongoClient(uri, **kwargs)
        intf = MongoDBAdapter(db)
    else:
        raise ValueError(f"protocol not supported: {protocol}")
    if not for_write:
        intf = NoSQLReadOnly(intf)
    return intf


class MongoDBAdapter(NoSqlDatabase):
    """
    Extremely simple interface to pymongo.
    """
    def __init__(self, db):
        self.db = db

    def query(self, collection: str, query: dict=None, limit: int=None, **kwargs) -> typing.Iterable[dict]:
        return self.db[collection].query(query, limit=limit, **kwargs)

    def insert(self, collection: str, doc: dict, **kwargs):
        resp = self.db[collection].insert_one(doc, **kwargs)
        return str(resp.inserted_id)

    def update(self, collection: str, filter: dict, changes: dict, upsert: bool=False, **kwargs):
        return self.db[collection].update_many(filter, update=changes, upsert=upsert, **kwargs).matched_count

    def delete(self, collection: str, filter: dict, **kwargs):
        return self.db[collection].delete_many(filter, **kwargs).modified_count



import os
import tempfile
import shutil
import unittest
import dslibrary
from dslibrary import ENV_DSLIBRARY_TARGET, ENV_DSLIBRARY_SPEC, run_model_method, DSLibrary
from dslibrary.transport.to_local import DSLibraryLocal
from dslibrary.transport.to_mmlibrary import DSLibraryViaMMLibrary
from dslibrary.transport.to_rest import DSLibraryViaREST
from dslibrary.transport.to_volume import DSLibraryViaVolume
from tests.t_utils import reset_env


class TestRunModelMethod(unittest.TestCase):

    def test_run_model_method(self):
        def m(a: int):
            return a+1
        dsl = DSLibrary(spec={"parameters": {"a": "4", "b": "b"}})
        r = run_model_method(m, dsl=dsl)
        assert r == 5


class TestMMFront(unittest.TestCase):

    def test_new_instance__local(self):
        try:
            tmpdir = tempfile.mkdtemp()
            os.environ[ENV_DSLIBRARY_TARGET] = f"local:{tmpdir}"
            mm = dslibrary.instance()
            assert isinstance(mm, DSLibraryLocal)
            assert mm._root == tmpdir
            shutil.rmtree(tmpdir)
        finally:
            os.environ[ENV_DSLIBRARY_TARGET] = ""

    def test_new_instance__rest(self):
        try:
            os.environ[ENV_DSLIBRARY_TARGET] = f"http://host:1234/path/"
            mm = dslibrary.instance()
            assert isinstance(mm, DSLibraryViaREST)
            assert mm._url == 'http://host:1234/path/'
        finally:
            os.environ[ENV_DSLIBRARY_TARGET] = ""

    def test_new_instance__mmlibrary(self):
        try:
            os.environ[ENV_DSLIBRARY_TARGET] = "mmlibrary"
            mm = dslibrary.instance()
            assert isinstance(mm, DSLibraryViaMMLibrary)
        finally:
            os.environ[ENV_DSLIBRARY_TARGET] = ""

    def test_new_instance__volume(self):
        try:
            os.environ[ENV_DSLIBRARY_TARGET] = f"volume:/path"
            mm = dslibrary.instance()
            assert isinstance(mm, DSLibraryViaVolume)
            assert mm._volume == "/path"
        finally:
            os.environ[ENV_DSLIBRARY_TARGET] = ""

    def tearDown(self) -> None:
        reset_env("test_pkg")



"""
Incoming REST calls can be sent here to be converted back into dslibrary calls.
"""
import json
import typing
import urllib.parse

from dslibrary import DSLibrary


class RequestInfo(object):
    """
    Use this class to describe the incoming request.
    """
    def __init__(self, method: str, path: str, post_data: (bytes, bytearray)=None, headers: dict=None):
        self.method = method.lower()
        if not path.startswith("/"):
            path = "/" + path
        if "?" in path:
            path, query = path.split("?", maxsplit=1)
        else:
            query = ""
        self.path = path
        query_args = urllib.parse.parse_qs(query, keep_blank_values=True, encoding="utf-8", errors="ignore")
        self.query_args = {k: "" if len(v) == 0 else v[0] for k, v in query_args.items()}
        self.post_data = post_data
        self.headers = {k.lower(): v for k, v in (headers or {}).items()}


class ResponseInfo(object):
    """
    An instance of this class describes a response.
    """
    def __init__(self, content: (bytes, bytearray, str, typing.IO, dict, list, tuple), headers: dict=None, status_code: int=200):
        self.content = content
        self.headers = headers or {}
        self.status_code = status_code

    def close(self):
        if hasattr(self.content, "close"):
            self.content.close()


class HttpToDsl(object):
    """
    An instance of this class translates requests into actions against 'target', and returns responses.

    Structure of a REST service:
      * receive all requests to a particular path
      * do authentication, i.e. by checking HTTP headers for a token or credentials
      * do any other needed processing of the HTTP headers, which might supply other important context
      * turn the inbound request into a RequestInfo() object
      * call the translate() method
      * send a response based on the returnedResponseInfo() object
      * call the close() method of the response object
    """
    def __init__(self, target: DSLibrary, base_path: str=""):
        self.target = target
        self.base_path = "/" + base_path.strip("/")
        if not self.base_path.endswith("/"):
            self.base_path += "/"

    @staticmethod
    def err_400(message: str):
        return ResponseInfo(f"bad request: {message}", status_code=404)

    @staticmethod
    def err_404():
        return ResponseInfo("not found", status_code=404)

    @staticmethod
    def ok():
        return ResponseInfo({})

    def translate(self, request: RequestInfo) -> ResponseInfo:
        """
        Main entry point for translation of HTTP requests.
        """
        if not request.path.startswith(self.base_path):
            return self.err_404()
        path = request.path[len(self.base_path):]
        path_parts = path.strip("/").split("/")
        method_name = f"_{request.method}_{path_parts[0]}"
        if not hasattr(self, method_name):
            return self.err_404()
        return getattr(self, method_name)(request)

    def _get_context(self, request: RequestInfo) -> ResponseInfo:
        """
        Get all the constant elements of our target: parameters, metadata, etc..
        """
        ctx = {
            "uri": self.target.get_uri(),
            "metadata": self.target.get_metadata().to_json(),
            "parameters": self.target.get_parameters()
        }
        return ResponseInfo(content=ctx, headers={"Content-Type": "application/json"})

    def _read_resources(self, request: RequestInfo, open_method: typing.Callable, name: str) -> ResponseInfo:
        # TODO use headers instead of arg for byte range
        byte_range = json.loads(request.query_args.get("byte_range") or 'null')
        with open_method(name, mode='rb') as f_r:
            if byte_range:
                f_r.seek(byte_range[0])
                content = f_r.read(byte_range[1] - byte_range[0])
            else:
                content = f_r.read()
        return ResponseInfo(content, {"Content-Type": "application/octet-stream"})

    def _write_resources(self, request: RequestInfo, open_method: typing.Callable, name: str) -> ResponseInfo:
        append = request.query_args.get("append")
        with open_method(name, mode='ab' if append else 'wb') as f_w:
            f_w.write(request.post_data)
        return self.ok()

    def _get_resources(self, request: RequestInfo) -> ResponseInfo:
        """
        Read part of a resource.
        """
        resource_name = request.query_args.get("resource_name")
        if not resource_name:
            return self.err_400("missing 'resource_name'")
        return self._read_resources(request, open_method=self.target.open_resource, name=resource_name)

    def _put_resources(self, request: RequestInfo) -> ResponseInfo:
        """
        Write or append to a resource.
        """
        resource_name = request.query_args.get("resource_name")
        if not resource_name:
            return self.err_400("missing 'resource_name'")
        return self._write_resources(request, open_method=self.target.open_resource, name=resource_name)

    def _get_run_data(self, request: RequestInfo) -> ResponseInfo:
        """
        Read run data.
        """
        resource_name = request.query_args.get("resource_name")
        if not resource_name:
            return self.err_400("missing 'resource_name'")
        return self._read_resources(request, open_method=self.target.open_run_data, name=resource_name)

    def _put_run_data(self, request: RequestInfo) -> ResponseInfo:
        """
        Write run data.
        """
        resource_name = request.query_args.get("resource_name")
        if not resource_name:
            return self.err_400("missing 'resource_name'")
        return self._write_resources(request, open_method=self.target.open_run_data, name=resource_name)

    def _get_model_binary(self, request: RequestInfo) -> ResponseInfo:
        """
        Read model binary file(s).
        """
        part = request.query_args.get("part")
        return self._read_resources(request, open_method=self.target.open_model_binary, name=part)

    def _put_model_binary(self, request: RequestInfo) -> ResponseInfo:
        """
        Write mode binary data.
        """
        part = request.query_args.get("part")
        return self._write_resources(request, open_method=self.target.open_model_binary, name=part)

    def _get_db(self, request: RequestInfo) -> ResponseInfo:
        """
        Read chunk of SQL results.
        """
        # TODO see data_access code

    def _post_db(self, request: RequestInfo) -> ResponseInfo:
        """
        Execute SQL.
        """
        # TODO see data_access code


    # TODO scoring-requests
    # TODO score
    # TODO filesystem functions

    # TODO Not implemented yet in to_rest:
    #   - nosql, filesystem


def tornado_to_dslibrary(target: DSLibrary, verifier: typing.Callable=None):
    """
    Generate a Tornado app that proxies data requests to a DSLibrary instance.

    This shortens creation of a data service endpoint to the following:
        app = tornado_to_dslibrary(target)
        app.listen(port)
        tornado.ioloop.IOLoop.current().start()

    :param target:   Data handler.
    :param verifier: Checks every request for validity, i.e. ensuring it has a valid access token.  It is passed a
        RequestInfo instance, and can return a DSLibrary instance.  It should raise a tornado.web.* exception if
        the authorization is incorrect.
    :return:    Tornado app object.
    """
    import tornado.web
    to_dsl = HttpToDsl(target)
    class MainHandler(tornado.web.RequestHandler):
        def run(self, rqst: RequestInfo):
            to_dsl = None
            if verifier:
                alt_dsl = verifier(rqst)
                if alt_dsl:
                    to_dsl = HttpToDsl(alt_dsl)
            try:
                resp = to_dsl.translate(rqst)
                for k, v in resp.headers.items():
                    self.set_header(k, v)
                self.write(resp.content)
            except Exception as err:
                raise tornado.web.HTTPError(500, f"Internal error: {err}")

        def _common(self):
            path = self.request.path
            if self.request.query:
                path += "?" + self.request.query
            return {
                "path": path, "headers": self.request.headers
            }

        async def get(self):
            rqst = RequestInfo(
                method="get", **self._common()
            )
            return self.run(rqst)

        async def post(self):
            rqst = RequestInfo(
                method="post", **self._common(), post_data=self.request.body
            )
            return self.run(rqst)

        async def put(self):
            rqst = RequestInfo(
                method="put", **self._common(), post_data=self.request.body
            )
            return self.run(rqst)

        async def delete(self):
            rqst = RequestInfo(
                method="delete", **self._common()
            )
            return self.run(rqst)

    return tornado.web.Application([
        (r"/.*", MainHandler),
    ])



import unittest
import mock
import os
import tempfile
import shutil

import dslibrary
from tests.t_utils import reset_env


class TestRunNoContext(unittest.TestCase):
    """
    When you use dslibrary with no context at all, it makes some simple assumptions.
    """

    def test_default_parameters_from_command_line(self):
        # --NAME=VALUE
        dsl = dslibrary.instance()
        with mock.patch("sys.argv", ["source.py", "--x=7"]):
            r = dsl.get_parameter("x")
            assert r == 7, r
        # --NAME VALUE
        dsl = dslibrary.instance()
        with mock.patch("sys.argv", ["source.py", "--x", "123"]):
            r = dsl.get_parameter("x")
            assert r == 123, r
        # string type
        dsl = dslibrary.instance()
        with mock.patch("sys.argv", ["source.py", "--x", "123a"]):
            r = dsl.get_parameter("x")
            assert r == "123a", r

    def test_open_resource_local_file_in_cwd(self):
        """
        All inputs and outputs are just files relative to the current directory.
        """
        project = tempfile.mkdtemp()
        with open(os.path.join(project, "f"), 'w') as f:
            f.write("abc")
        cwd0 = os.getcwd()
        os.chdir(project)
        try:
            dsl = dslibrary.instance()
            r = dsl.read_resource("f", mode="r")
            assert r == "abc"
            r = dsl.read_resource("f", mode="rb")
            assert r == b"abc"
        finally:
            os.chdir(cwd0)
        shutil.rmtree(project)

    def tearDown(self) -> None:
        reset_env("test_run_no_context")
